/*
 **  CWFolder.m
 **
 **  Copyright (c) 2001-2006
 **
 **  Author: Ludovic Marcotte <ludovic@Sophos.ca>
 **
 **  This library is free software; you can redistribute it and/or
 **  modify it under the terms of the GNU Lesser General Public
 **  License as published by the Free Software Foundation; either
 **  version 2.1 of the License, or (at your option) any later version.
 **  
 **  This library is distributed in the hope that it will be useful,
 **  but WITHOUT ANY WARRANTY; without even the implied warranty of
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 **  Lesser General Public License for more details.
 **  
 **  You should have received a copy of the GNU Lesser General Public
 **  License along with this library; if not, write to the Free Software
 **  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <Pantomime/CWFolder.h>

#include <Pantomime/CWConstants.h>
#include <Pantomime/CWContainer.h>
#include <Pantomime/CWFlags.h>
#include <Pantomime/CWMessage.h>
#include <Pantomime/NSString+Extensions.h>

#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSMapTable.h>

#include "CWDummyMessage.h"
//
//
//
@implementation CWFolder 

- (id) initWithName: (NSString *) theName
{
    self = [super init];
    
    _properties = [[NSMutableDictionary alloc] init];
    _allVisibleMessages = nil;
    
    allMessages = [[NSMutableArray alloc] init];
    
    //
    // By default, we don't do message threading so we don't
    // initialize this ivar for no reasons
    //
    _threadContainer = nil;
    _cacheManager = nil;
    _mode = PantomimeUnknownMode;
    
    [self setName: theName];
    [self setShowDeleted: NO];
    [self setShowRead: YES];
    
    return self;
}


//
//
//
- (void) dealloc
{
    //NSLog(@"Folder: -dealloc");
    RELEASE(_properties);
    RELEASE(_name);
    if (_threadContainer)
        [self unthread];
    TEST_RELEASE(_threadContainer);
    
    //
    // To be safe, we set the value of the _folder ivar of all CWMessage
    // instances to nil value in case something is retaining them.
    //
    [allMessages makeObjectsPerformSelector: @selector(setFolder:) withObject: nil];
    RELEASE(allMessages);
    
    TEST_RELEASE(_allVisibleMessages);
    TEST_RELEASE(_cacheManager);
    
    [super dealloc];
}


//
// NSCopying protocol (FIXME)
//
- (id) copyWithZone: (NSZone *) zone
{
    return RETAIN(self);
}


//
//
//
- (NSString *) name
{
    return _name;
}


//
//
//
- (void) setName: (NSString *) theName
{
    ASSIGN(_name, theName);
}


//
//
//
- (void) appendMessage: (CWMessage *) theMessage
{
    if (theMessage)
    {
        [allMessages addObject: theMessage];
        
        if (_allVisibleMessages)
        {
            [_allVisibleMessages addObject: theMessage];
        }
        
    }
}


//
//
//
- (void) appendMessageFromRawSource: (NSData *) theData
                              flags: (CWFlags *) theFlags
{
    [self subclassResponsibility: _cmd];
}


//
//
//
- (NSArray *) allMessages
{ 
    if (_allVisibleMessages == nil)
    {
        int i, count;
        
        count = [allMessages count];
        _allVisibleMessages = [[NSMutableArray alloc] initWithCapacity: count];
        
        // quick
        if (_show_deleted && _show_read)
        {
            [_allVisibleMessages addObjectsFromArray: allMessages];
            return _allVisibleMessages;
        }
        
        for (i = 0; i < count; i++)
        {
            CWMessage *aMessage;
            
            aMessage = [allMessages objectAtIndex: i];
            
            // We show or hide deleted messages
            if (_show_deleted)
            {
                [_allVisibleMessages addObject: aMessage];
            }
            else
            {
                if ([[aMessage flags] contain: PantomimeDeleted])
                {
                    // Do nothing
                    continue;
                }
                else
                {
                    [_allVisibleMessages addObject: aMessage];
                }
            }
            
            // We show or hide read messages
            if (_show_read)
            {
                if (![_allVisibleMessages containsObject: aMessage])
                {
                    [_allVisibleMessages addObject: aMessage];
                }
            }
            else
            {
                if ([[aMessage flags] contain: PantomimeSeen])
                {
                    if (![[aMessage flags] contain: PantomimeDeleted])
                    {
                        [_allVisibleMessages removeObject: aMessage];
                    }
                }
                else if (![_allVisibleMessages containsObject: aMessage])
                {
                    [_allVisibleMessages addObject: aMessage];
                }
            }
        }
    }
    
    return _allVisibleMessages;
}


//
//
//
- (void) setMessages: (NSArray *) theMessages
{
    if (theMessages)
    {
        RELEASE(allMessages);
        allMessages = [[NSMutableArray alloc] initWithArray: theMessages];
        
        if ([self threadContainer])
        {
            [self thread];
        }
    }
    else
    {
        DESTROY(allMessages);
    }
    
    DESTROY(_allVisibleMessages);
}


//
//
//
- (CWMessage *) messageAtIndex: (int) theIndex
{
    if (theIndex < 0 || theIndex >= [self count])
    {
        return nil;
    }
    
    return [[self allMessages] objectAtIndex: theIndex];
}


//
//
//
- (int) count
{
    return [[self allMessages] count];
}


//
//
//
- (void) close
{
    [self subclassResponsibility: _cmd];
    return;
}


//
//
//
- (void) expunge
{
    [self subclassResponsibility: _cmd];
}


//
//
//
- (id) store
{
    return _store;
}


//
// No need to retain the store here since our store object
// retains our folder object.
//
- (void) setStore: (id) theStore
{
    _store = theStore;
}


//
//
//
- (void) removeMessage: (CWMessage *) theMessage
{
    if (theMessage)
    {
        [allMessages removeObject: theMessage];
        
        if (_allVisibleMessages)
        {
            [_allVisibleMessages removeObject: theMessage];
        }
        
        // FIXME - We must go through our _allContainers ivar in order
        //         to find the message that has just been removed from
        //         this folder. We must go through all levels.
        //         Right now, we simply do again our message threading algo
        if ([self threadContainer])
        {
            [self thread];
        }
    }
}


//
//
//
- (BOOL) showDeleted
{
    return _show_deleted;
}


//
//
//
- (void) setShowDeleted: (BOOL) theBOOL
{
    if (theBOOL != _show_deleted)
    {
        _show_deleted = theBOOL;
        DESTROY(_allVisibleMessages);
    }
}


//
//
//
- (BOOL) showRead
{
    return _show_read;
}


//
//
//
- (void) setShowRead: (BOOL) theBOOL
{
    if (theBOOL != _show_read)
    {
        _show_read = theBOOL;
        DESTROY(_allVisibleMessages);
    }
}


//
//
//
- (int) numberOfDeletedMessages
{
    int c, i, count;
    
    c = [allMessages count];
    count = 0;
    
    for (i = 0; i < c; i++)
    {
        if ([[[allMessages objectAtIndex: i] flags] contain: PantomimeDeleted])
        {
            count++;
        }
    }
    
    return count;
}


//
//
//
- (int) numberOfUnreadMessages
{
    int i, c, count;
    
    c = [allMessages count];
    count = 0;
    
    for (i = 0; i < c; i++)
    {
        if (![[[allMessages objectAtIndex: i] flags] contain: PantomimeSeen])
        {
            count++;
        }
    }
    
    return count;
}


//
//
//
- (long) size;
{
    long size;
    int c, i;
    
    c = [allMessages count];
    size = 0;
    
    for (i = 0; i < c; i++)
    {
        size += [(CWMessage *)[allMessages objectAtIndex: i] size];
    }
    
    return size;
    
}


//
//
//
- (void) updateCache
{
    DESTROY(_allVisibleMessages);
}

#if 0
//
//
//
- (void) thread
{
    NSMapTable *id_table, *subject_table;
    NSMutableArray *referenceArray;
    NSAutoreleasePool *pool;
    int i, count;
    BOOL hasValidDelegate;
    
    hasValidDelegate = [[self delegate] respondsToSelector: @selector(missingMessageWithID:referencedInMessage:)];
    
    // We clean up ...
    TEST_RELEASE(_allContainers);
    
    // We create our local autorelease pool
    pool = [[NSAutoreleasePool alloc] init];
    
    // Build id_table and our containers mutable array
    id_table = NSCreateMapTable(NSObjectMapKeyCallBacks, NSObjectMapValueCallBacks, 16);
    _allContainers = [[NSMutableArray alloc] initWithCapacity: 20];
    referenceArray = [[NSMutableArray alloc] initWithCapacity: 20];
    //
    // 1. A., B. and C.
    //
    count = [allMessages count];
    for (i = 0; i < count; i++)
    {
        CWContainer *aContainer;
        CWMessage *aMessage;
        
        NSString *aReference;
        int j;
        
        // So that gcc shutup
        aMessage = nil;
        aReference = nil;
        
        aMessage = [allMessages objectAtIndex: i];
        
        // We skip messages that don't have a valid Message-ID
        if (![aMessage messageID])
        {
            aContainer = [[CWContainer alloc] init];
            aContainer->message = aMessage;
            [aMessage setProperty: aContainer  forKey: @"Container"];
            [_allContainers addObject: aContainer];
            RELEASE(aContainer);
            continue;
        }
        
        //
        // A.
        //
        aContainer = NSMapGet(id_table, [aMessage messageID]);
        
        if (aContainer)
        {
            //aContainer->message = aMessage;
            
            if (aContainer->message != aMessage)
            {
                aContainer = [[CWContainer alloc] init];
                aContainer->message = aMessage;
                [aMessage setProperty: aContainer  forKey: @"Container"];
                NSMapInsert(id_table, [aMessage messageID], aContainer);
                DESTROY(aContainer);
            }
        }
        else
        {
            aContainer = [[CWContainer alloc] init];
            aContainer->message = aMessage;
            [aMessage setProperty: aContainer  forKey: @"Container"];
            NSMapInsert(id_table, [aMessage messageID], aContainer);
            DESTROY(aContainer);
        }
        
        //
        // B. For each element in the message's References field:
        //
        [referenceArray addObjectsFromArray: [aMessage allReferences]];
        if ([aMessage inReplyTo] && ! [referenceArray containsObject: [aMessage inReplyTo]])
            [referenceArray insertObject: [aMessage inReplyTo] atIndex: 0];
        for (j = 0; j < [referenceArray count]; j++)
        {
            // We get a Message-ID
            aReference = [referenceArray objectAtIndex: j];
            
            // Find a container object for the given Message-ID
            aContainer = NSMapGet(id_table, aReference);
            
            if (aContainer)
            {
                // We found it. We use that.
            }
            // Otherwise, make (and index) one (new Container) with a null Message
            else 
            {
                aContainer = [[CWContainer alloc] init];
                if (hasValidDelegate)
                {
                    // Try if there is a Sent message with this messageID...
                    CWMessage *sentMsg;
                    
                    sentMsg = [[self delegate] performSelector: @selector(missingMessageWithID:referencedInMessage:)
                                                    withObject: aReference
                                                    withObject: aMessage];
                    if (sentMsg)
                        aContainer->message = sentMsg;
                }
                NSMapInsert(id_table, aReference, aContainer);
                RELEASE(aContainer);
            }
            
            // NOTE:
            // aContainer is valid here. It points to the message (could be a nil message)
            // that has a Message-ID equals to the current aReference value.
            
            // If we are currently using the last References's entry of our list,
            // we simply break the loop since we are gonna set it in C.
            //if ( j == ([[aMessage allReferences] count] - 1) )
            //  {
            //    break;
            // }
            
            // The last references
            if (j == ([referenceArray count] - 1) &&x
                aContainer->parent == nil)
            {
                // We grab the container of our current message
                [((CWContainer *)NSMapGet(id_table, [aMessage messageID])) setParent: aContainer];
            }
            
            // We set the child
            //if ( aContainer->message != aMessage &&
            //     aContainer->child == nil )
            //  {
            //    [aContainer setChild: NSMapGet(id_table, [aMessage messageID])];
            //  }	      
            
        } // for (j = 0; ...
        [referenceArray removeAllObjects];
        
        // NOTE: The loop is over here. It was an ascending loop so
        //       aReference points to the LAST reference in our References list
        
        //
        // C. Set the parent of this message to be the last element in References. 
        //
        // NOTE: Again, aReference points to the last Message-ID in the References list
        
        // We get the container for the CURRENT message
        aContainer = (CWContainer *)NSMapGet(id_table, [aMessage messageID]);
        
        // If we have no References and no In-Reply-To fields, we simply set a
        // the parent to nil since it can be the message that started the thread.
        CWContainer *oldParent;
        
        oldParent = aContainer->parent;
        if (oldParent)
            [oldParent setChild: nil];
        if ([[aMessage allReferences] count] == 0 && [aMessage inReplyTo] == nil)
        {
            [aContainer setParent: nil];
        }
        // If we have no References but an In-Reply-To field, that becomes our parent.
        else if ([[aMessage allReferences] count] == 0 && [aMessage inReplyTo])
        {
            CWContainer *inReplyToContainer;
            
            inReplyToContainer = (CWContainer *)NSMapGet(id_table, [aMessage inReplyTo]);
            [aContainer setParent: inReplyToContainer];
            // FIXME, should we really do that? or should we do it in B?
            [inReplyToContainer setChild: aContainer];
        }
        else
        {
            CWContainer *refContainer;
            
            refContainer = (CWContainer *)NSMapGet(id_table, aReference);
            [aContainer setParent: refContainer];
            [refContainer setChild: aContainer];
        }
        
    } // for (i = 0; ...
    
    //
    // 2. Find the root set.
    //
    [_allContainers addObjectsFromArray: NSAllMapTableValues(id_table)];
    
    //while (NO)
    for (i = ([_allContainers count] - 1); i >= 0; i--)
    {
        CWContainer *aContainer;
        
        aContainer = [_allContainers objectAtIndex: i];
        
        if (aContainer->parent != nil)
        {
            [_allContainers removeObjectAtIndex: i];
        }
    }
    
    //
    // 3. Discard id_table.
    //
    NSFreeMapTable(id_table);
    
    
    //
    // 4. Prune empty containers.
    //
    //while (NO)
    for (i = ([_allContainers count] - 1); i >= 0; i--)
    {
        CWContainer *aContainer;
        
        aContainer = [_allContainers objectAtIndex: i];
        
        // Recursively walk all containers under the root set.
        while (aContainer)
        {
            // A. If it is an empty container with no children, nuke it
            if (aContainer->message == nil &&
                aContainer->child == nil)
            {
                // We nuke it
                // FIXME: Won't work for non-root containers.
                [_allContainers removeObject: aContainer];
            }
            
            // B. If the Container has no Message, but does have children, remove this container but 
            //    promote its children to this level (that is, splice them in to the current child list.)
            //    Do not promote the children if doing so would promote them to the root set 
            //    -- unless there is only one child, in which case, do. 
            // FIXME: We promote to the root no matter what :)
            if (aContainer->message == nil && aContainer->child)
            {
                CWContainer *c;
                
                c = aContainer;
                RETAIN(c);
                [c->child setParent: nil];
                [_allContainers removeObject: c];
                [_allContainers addObject: c->child]; // We promote the the root for now
                
                // We go to our child and we continue to loop
                //aContainer = aContainer->child;
                aContainer = [aContainer childAtIndex: ([aContainer count]-1)];
                RELEASE(c);
                continue;
            }
            
            //aContainer = aContainer->child;
            aContainer = [aContainer childAtIndex: ([aContainer count]-1)];
        }
        
    }
    
    //
    // 5. Group root set by subject.
    //
    // A. Construct a new hash table, subject_table, which associates subject 
    //    strings with Container objects.
    subject_table = NSCreateMapTable(NSObjectMapKeyCallBacks, NSObjectMapValueCallBacks, 16);
    
    //
    // B. For each Container in the root set:
    //
    
    //while (NO)
    for (i = 0; i < [_allContainers count]; i++)
    {
        CWContainer *aContainer;
        CWMessage *aMessage;
        NSString *aString;
        
        aContainer = [_allContainers objectAtIndex: i];
        aMessage = aContainer->message;
        aString = [aMessage subject];
        
        if (aString)
        {
            aString = [aMessage baseSubject];
            
            // If the subject is now "", give up on this Container.
            if ([aString length] == 0)
            {
                //aContainer = aContainer->child;
                continue;
            }
            
            // We set the new subject
            //[aMessage setSubject: aString];
            
            // Add this Container to the subject_table if:
            // o There is no container in the table with this subject, or
            // o This one is an empty container and the old one is not: 
            //   the empty one is more interesting as a root, so put it in the table instead.
            // o The container in the table has a ``Re:'' version of this subject, 
            //   and this container has a non-``Re:'' version of this subject. 
            //   The non-re version is the more interesting of the two.
            if (!NSMapGet(subject_table, aString))
            {
                NSMapInsert(subject_table, aString, aContainer);
            }
            else
            {
                NSString *aSubject;
                
                // We obtain the subject of the message of our container.
                aSubject = [((CWContainer *)NSMapGet(subject_table, aString))->message subject];
                
                if ([aSubject hasREPrefix] && ![[aMessage subject] hasREPrefix])
                {
                    // We replace the container
                    NSMapRemove(subject_table, aString);
                    NSMapInsert(subject_table, [aMessage subject], aContainer);
                }
            }
            
        } // if ( aString )
    }
    
    //
    // C. Now the subject_table is populated with one entry for each subject which occurs in 
    //    the root set. Now iterate over the root set, and gather together the difference.
    //
    //while (NO)
    for (i = ([_allContainers count]-1); i >= 0; i--)
    {
        CWContainer *aContainer, *containerFromTable;
        NSString *aSubject, *aString;
        
        aContainer = [_allContainers objectAtIndex: i];
        
        // Find the subject of this Container (as above.)
        aSubject = [aContainer->message subject];
        aString = [aContainer->message baseSubject];
        
        // Look up the Container of that subject in the table.
        // If it is null, or if it is this container, continue.
        containerFromTable = NSMapGet(subject_table, aString);
        if (!containerFromTable || containerFromTable == aContainer) 
        {
            continue; 
        }
        
        // If that container is a non-empty, and that message's subject does 
        // not begin with ``Re:'', but this message's subject does, then make this be a child of the other.
        if (![[containerFromTable->message subject] hasREPrefix] &&
            [aSubject hasREPrefix])
        {
            [aContainer setParent: containerFromTable];
            [containerFromTable setChild: aContainer]; 
            [_allContainers removeObject: aContainer];
        }
        // If that container is a non-empty, and that message's subject begins with ``Re:'', 
        // but this  message's subject does not, then make that be a child of this one -- 
        // they were misordered. (This happens somewhat implicitly, since if there are two
        // messages, one with Re: and one without, the one without will be in the hash table,
        // regardless of the order in which they were seen.)
        else if ([[containerFromTable->message subject] hasREPrefix] &&
                 ![aSubject hasREPrefix])
        {
            [containerFromTable setParent: aContainer];
            [aContainer setChild: containerFromTable]; 
            [_allContainers removeObject: containerFromTable];
        }
        // Otherwise, make a new empty container and make both msgs be a child of it. 
        // This catches the both-are-replies and neither-are-replies cases, and makes them 
        // be siblings instead of asserting a hierarchical relationship which might not be true.
        else
        {
            // FIXME - not so sure about that step.
            CWContainer *aNewContainer;
            
            aNewContainer = [[CWContainer alloc] init];
            
            [aContainer setParent: aNewContainer];
            [containerFromTable setParent: aNewContainer];
            
            [aNewContainer setChild: aContainer]; 
            [aNewContainer setChild: containerFromTable];
            [_allContainers addObject: aNewContainer];
            RELEASE(aNewContainer);
            
            // We remove ..
            [_allContainers removeObject: aContainer];
            [_allContainers removeObject: containerFromTable];
        }
    }
    
    NSFreeMapTable(subject_table);
    
    //
    // 6.  Now you're done threading!
    //
    //     Specifically, you no longer need the ``parent'' slot of the Container object, 
    //     so if you wanted to flush the data out into a smaller, longer-lived structure, you 
    //     could reclaim some storage as a result. 
    //
    // GNUMail.app DOES USE the parent slot so we keep it.
    
    //
    // 7.  Now, sort the siblings.
    //     
    //     At this point, the parent-child relationships are set. However, the sibling ordering 
    //     has not been adjusted, so now is the time to walk the tree one last time and order the siblings 
    //     by date, sender, subject, or whatever. This step could also be merged in to the end of step 4, 
    //     above, but it's probably clearer to make it be a final pass. If you were careful, you could 
    //     also sort the messages first and take care in the above algorithm to not perturb the ordering,
    //      but that doesn't really save anything. 
    //
    // By default we at least sort everything by number.
    //[_allContainers sortUsingSelector: @selector(compareAccordingToNumber:)];
    DESTROY(referenceArray);
    RELEASE(pool);
}

//
//
//
- (void) unthread
{
    int count;
    
    count = [allMessages count];
    
    while (count--)
    {
        [(CWMessage *)[allMessages objectAtIndex: count] setProperty: nil  forKey: @"Container"];
    }
    
    DESTROY(_threadContainer);
}
#endif

//
//
//
- (void) search: (NSString *) theString
           mask: (PantomimeSearchMask) theMask
        options: (PantomimeSearchOption) theOptions
{
    [self subclassResponsibility: _cmd];
}


//
//
//
- (id) cacheManager
{
    return _cacheManager;
}

- (void) setCacheManager: (id) theCacheManager
{
    ASSIGN(_cacheManager, theCacheManager);
}


//
//
//
- (PantomimeFolderMode) mode
{
    return _mode;
}


//
//
//
- (void) setMode: (PantomimeFolderMode) theMode
{
    _mode = theMode;
}


//
//
//
- (void) setFlags: (CWFlags *) theFlags
         messages: (NSArray *) theMessages
{
    int c, i;
    
    c = [theMessages count];
    for (i = 0; i < c; i++)
    {
        [[theMessages objectAtIndex: i] setFlags: theFlags];
    }
}


//
//
//
- (id) propertyForKey: (id) theKey
{
    return [_properties objectForKey: theKey];
}


//
//
//
- (void) setProperty: (id) theProperty
              forKey: (id) theKey
{
    if (theProperty)
    {
        [_properties setObject: theProperty  forKey: theKey];
    }
    else
    {
        [_properties removeObjectForKey: theKey];
    }
}


//
// Added by Annard Brouwer
//
- (id)delegate
{
    return _delegate;
}

- (void)setDelegate:(id)aDelegate
{
    _delegate = aDelegate;
}

- (void)setThreadContainer:(CWContainer *)aContainer
{
    if (aContainer)
    {
        ASSIGN(_threadContainer, aContainer);
    }
    else
    {
        DESTROY(_threadContainer);
    }
}

- (CWContainer *)threadContainer
{
    return _threadContainer;
}

- (NSArray *)allContainers
{
    return [NSArray arrayWithObject: [self threadContainer]];
}

/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * This is an implementation of a message threading algorithm, as originally devised by Zamie Zawinski.
 * See <a href="http://www.jwz.org/doc/threading.html">http://www.jwz.org/doc/threading.html</a> for details.
 * For his Java implementation, see <a href="http://lxr.mozilla.org/mozilla/source/grendel/sources/grendel/view/Threader.java">http://lxr.mozilla.org/mozilla/source/grendel/sources/grendel/view/Threader.java</a>
 *  
 * @author rwinston <rwinston@checkfree.com>
 * Converted to Objective-C by Annard
 *
 */

- (CWMessage *)createDummyMessage
{
    CWMessage *dummyMsg;

    dummyMsg = [[CWDummyMessage alloc] init];
    [_dummyMessageArray addObject: dummyMsg];
    [dummyMsg release];
    return dummyMsg;
}

/**
 * The main threader entry point - The client passes in an array of CWMessage objects, and 
 * this method constructs a connected 'graph' of messages
 */
- (void)thread
{
    NSAutoreleasePool *pool;
    NSMutableDictionary *idTable;
    int count, i;
    CWContainer *rootContainer, *c;
    
    count = [allMessages count];
    if (0 == count)
        return;

    if (nil == _dummyMessageArray)
        _dummyMessageArray = [[NSMutableArray alloc] initWithCapacity: 20];
    
    // We create our local autorelease pool
    pool = [[NSAutoreleasePool alloc] init];
    
    idTable = [NSMutableDictionary dictionaryWithCapacity: count];
    
    // walk through each CWMessage *element
    for (i = 0; i < count; i++)
    {
        CWMessage *aMsg;
        
        aMsg = [allMessages objectAtIndex: i];
        if ([[aMsg subject] length] > 0)
            [self buildContainer: aMsg withMessageIdTable: idTable];
    }
    
    rootContainer = [self findRootSetUsingMessageIDTable: idTable];
    [idTable removeAllObjects];
    
    [self pruneEmptyContainers: rootContainer];
    
    [rootContainer reverseChildren];
    [self gatherSubjectsOfContainer: rootContainer];
    
    if ([rootContainer next] != nil)
        [NSException raise: NSInvalidArgumentException
                    format: @"rootContainer node has a next: %@", rootContainer];
                    
    for (c = [rootContainer child]; c != nil; c = [c next])
    {
        if ([c message] == nil)
            [c setMessage: [self createDummyMessage]];
    }
    
    [rootContainer retain];
    RELEASE(pool);
    [self setThreadContainer: rootContainer];
    [rootContainer release];
}

- (void)unthread
{
    int count;
    
    count = [allMessages count];
    while (count-- > 0)
    {
        [(CWMessage *)[allMessages objectAtIndex: count] setProperty: nil forKey: @"Container"];
    }
    count = [_dummyMessageArray count];
    while (count-- > 0)
    {
        [(CWMessage *)[_dummyMessageArray objectAtIndex: count] setProperty: nil forKey: @"Container"];
    }
    DESTROY(_dummyMessageArray);
    [self setThreadContainer: nil];
}

- (void)buildContainer:(CWMessage *)msg
    withMessageIdTable:(NSMutableDictionary *)idTable
{
    NSString *msgID = [msg messageID];
    CWContainer *container = [idTable objectForKey: msgID];
    
    // A CWContainer exists for this msgID already. This should be a forward reference, but may 
    // be a duplicate msgID, in which case we will need to generate a bogus placeholder msgID
    if (container != nil)
    {
        if ([container message] != nil)
        {
            // oops! duplicate ids...
            msgID = [NSString stringWithFormat: @"<Bogus-msgID:%i>", bogusIdCount++];
            container = nil;
        }
        else
        {
            // The container just contained a forward reference to this message, so let's
            // fill in the msg field of the container with this message
            [container setMessage: msg];
        }
    }
    
    // No container exists for that message-id. Create one and insert it into the hash table.
    if (container == nil)
    {
        container = [CWContainer newContainer];
        [container setMessage: msg];
        [idTable setObject: container forKey: msgID];
    }
    
    // Iterate through all of the references and create CWContainers for any references that
    // don't have them.
    CWContainer *parentRef = nil;
    {
        NSMutableArray *references;
        unsigned i;
        BOOL hasValidDelegate;

        references = [NSMutableArray arrayWithArray: [msg allReferences]];
        // Be sure to add the In-Reply-To header if necessary
        if ([msg inReplyTo] && ! [references containsObject: [msg inReplyTo]])
            [references addObject: [msg inReplyTo]];

        hasValidDelegate = [[self delegate] respondsToSelector: @selector(missingMessageWithID:referencedInMessage:)];
        for (i = 0; i < [references count]; i++)
        {
            NSString *refString = [references objectAtIndex: i];
            CWContainer *ref = [idTable objectForKey: refString];
            
            // if this msgID doesnt have a container, create one
            if (ref == nil)
            {
                ref = [CWContainer newContainer];
                if (hasValidDelegate)
                {
                    // Try if there is a Sent message with this messageID...
                    CWMessage *sentMsg;
                    
                    sentMsg = [[self delegate] performSelector: @selector(missingMessageWithID:referencedInMessage:)
                                                    withObject: refString
                                                    withObject: msg];
                    if (sentMsg)
                        [ref setMessage: sentMsg];
                }
                [idTable setObject: ref forKey: refString];
            }
            
            // Link references together in the order they appear in the References: header,
            // IF they dont have a have a parent already &&
            // IF it will not cause a circular reference
            if ((parentRef != nil)
                && ([ref parent] == nil)
                && (parentRef != ref)
                && (! [parentRef findChild: ref])
                && (! [ref findChild: parentRef]))
            {
                // Link ref into the parent's child list
                [ref setParent: parentRef];
                [ref setNext: [parentRef child]];
                [parentRef setChild: ref];
            }
            parentRef = ref;
        }
    }
    
    // parentRef is now set to the container of the last element in the references field. make that 
    // be the parent of this container, unless doing so causes a circular reference
    if (parentRef != nil
        && (parentRef == container || [container findChild: parentRef]))
        parentRef = nil;
    
    // if it has a parent already, its because we saw this message in a References: field, and presumed
    // a parent based on the other entries in that field. Now that we have the actual message, we can
    // throw away the old parent and use this new one
    if ([container parent] != nil)
    {
        CWContainer *rest, *prev;
        
        for (prev = nil, rest = [[container parent] child];
             rest != nil;
             prev = rest, rest = [rest next])
        {
            if (rest == container)
                break;
        }
        
        if (rest == nil)
        {
            [NSException raise: NSInvalidArgumentException
                        format: @"Didn't find %@ in parent %@", container, [container parent]];
        }
        
        // Unlink this container from the parent's child list
        if (prev == nil)
            [[container parent] setChild: [container next]];
        else
            [prev setNext: [container next]];
        
        [container setNext: nil];
        [container setParent: nil];
    }
    
    // If we have a parent, link container into the parents child list
    if (parentRef != nil)
    {
        [container setParent: parentRef];
        [container setNext: [parentRef child]];
        [parentRef setChild: container];
    }
}

/**
 * Find the root set of all existing ThreadContainers
 * @return root the CWContainer representing the root node
 */
- (CWContainer *)findRootSetUsingMessageIDTable:(NSMutableDictionary *)idTable
{
    CWContainer *newRoot = [CWContainer newContainer];
    NSEnumerator *iter = [idTable keyEnumerator];
    NSString *anID;
    
    while (anID = [iter nextObject])
    {
        CWContainer *c = [idTable objectForKey: anID];
        if ([c parent] == nil)
        {
            if ([c next] != nil)
                [NSException raise: NSInvalidArgumentException
                            format: @"[c next] is not nil but %@", [c next]];
            [c setNext: [newRoot child]];
            [newRoot setChild: c];
        }
    }
    return newRoot;
}

/**
 * Delete any empty or dummy ThreadContainers
 * @param parent
 */
- (void)pruneEmptyContainers:(CWContainer *)parent
{
    CWContainer *container, *prev, *next;
    
    for (prev = nil, container = [parent child], next = [container next];
         container != nil;
         prev = container, container = next, next = (container == nil ? nil : [container next]))
    {
        // Is it empty and without any children? If so,delete it 
        if ([container message] == nil && [container child] == nil)
        {
            if (prev == nil)
                [parent setChild: [container next]];
            else
                [prev setNext: [container next]];

            // Resolve leaks
            [container setParent: nil];
            [container setNext: nil];
            // Set container to prev so that prev keeps its same value the next time through the loop
            container = prev;
        }
        // Else if empty, with kids, and (not at root or only one kid)
        else if ([container message] == nil && container.child != nil
                 && ([container parent] != nil || [[container child] next] == nil))
        {
            // We have an invalid/expired message with kids. Promote the kids to this level. 
            CWContainer *tail;
            CWContainer *kids = [container child];
            
            // Remove this container and replace with 'kids'.
            if (prev == nil)
                [parent setChild: kids];
            else
                [prev setNext: kids];
            
            // Make each child's parent be this level's parent -> i.e. promote the children. Make the last child's next point to this container's next
            // i.e. splice kids into the list in place of container
            for (tail = kids; [tail next] != nil; tail = [tail next])
                [tail setParent: [container parent]];
            
            [tail setParent: [container parent]];
            [tail setNext: [container next]];
            
            // next currently points to the item after the inserted items in the chain - reset that so we process the newly
            // promoted items next time round
            next = kids;

            // Resolve leaks
            [container setParent: nil];
            [container setChild: nil];
            [container setNext: nil];

            // Set container to prev so that prev keeps its same value the next time through the loop
            container = prev;
        }
        else if ([container child] != nil)
        {
            // A real message , with kids
            // Iterate over the children
            [self pruneEmptyContainers: container];
        }
    }
}

/**
 *  If any two members of the root set have the same subject, merge them. This is to attempt to accomodate messages without References: headers. 
 */
- (void)gatherSubjectsOfContainer:(CWContainer *)root
{
    CWContainer *c;
    int count = 0;
    NSMutableDictionary *subjectTable;
    
    for (c = [root child]; c != nil; c = [c next])
        count++;
    
    subjectTable = [NSMutableDictionary dictionaryWithCapacity: (unsigned)ceil(count * 1.2)];
    count = 0;
    
    for (c = [root child]; c != nil; c = [c next])
    {
        CWMessage *msg = [c message];
        
        // No msg? If so, it is a dummy node in the root set.
        // Only root set members may be dummies, and they alway have at least 2 kids
        // Take the first kid as representative of the subject
        if (msg == nil)
            msg = [[c child] message];
        
        NSString *subj = [msg baseSubject];
        
        if (subj == nil || [subj length] == 0)
            continue;
        
        CWContainer *old = [subjectTable objectForKey: subj];
        
        // Add this container to the table iff:
        // - There exists no container with this subject
        // - or this is a dummy container and the old one is not
        // - the dummy one is more interesting as a root, so put it in the table instead
        // - The container in the table has a "Re:" version of this subject, and 
        // this container has a non-"Re:" version of this subject. The non-"Re:" version
        // is the more interesting of the two.
        if (old == nil
            || ([c message] == nil && [old message] != nil)
            || ([old message] != nil
                && [[old message] subjectIsReply]
                && [c message] != nil
                && ! [[c message] subjectIsReply]))
        {
            [subjectTable setObject: c forKey: subj];
            count++;
        }
    }
    
    // If the table is empty, we're done
    if (count == 0)
        return;
    
    // subjectTable is now populated with one entry for each subject which occurs in the 
    // root set. Iterate over the root set, and gather together the difference.
    CWContainer *prev, *rest;

    for (prev = nil, c = [root child], rest = [c next];
         c != nil;
         prev = c, c = rest, rest = (rest == nil ? nil : [rest next]))
    {
        CWMessage *msg = [c message];
        
        // is it a dummy node?
        if (msg == nil)
            msg = [[c child] message];
        
        NSString *subj = [msg baseSubject];
        
        // Dont thread together all subjectless messages
        if (subj == nil || [subj length] == 0)
            continue;
        
        CWContainer *old = [subjectTable objectForKey: subj];
        
        if (old == c) // That's us
            continue;
        
        // We have now found another container in the root set with the same subject
        // Remove the "second" message from the root set
        if (prev == nil)
            [root setChild: [c next]];
        else
            [prev setNext: [c next]];
        [c setNext: nil];
        
        if ([old message] == nil && [c message] == nil)
        {
            // both dummies - merge them
            CWContainer *tail;
            
            for (tail = [old child];
                 tail != nil && [tail next] != nil;
                 tail = [tail next]);
            
            [tail setNext: [c child]];
            
            for (tail = [c child]; tail != nil; tail = [tail next])
                [tail setParent: old];

            // Resolve leaks
            [c setParent: nil];
            [c setNext: nil];
            [c setChild: nil];
        }
        else if ([old message] == nil
                 || ([c message] != nil
                     && [[c message] subjectIsReply]
                     && ! [[old message] subjectIsReply]))
        {
            // Else if old is empty, or c has "Re:" and old does not  ==> make this message a child of old
            [c setParent: old];
            [c setNext: [old child]];
            [old setChild: c];
        }
        else
        {
            // else make the old and new messages be children of a new dummy container.
            // We create a new container object for old.msg and empty the old container
            CWContainer *newc = [CWContainer newContainer];
            CWContainer *tail;
            
            [newc setMessage: [old message]];
            [newc setChild: [old child]];
            
            for (tail = [newc child];
                 tail != nil;
                 tail = [tail next])
                [tail setParent: newc];
            
            [old setMessage: nil];
            [old setChild: nil];
            
            [c setParent: old];
            [newc setParent: old];
            
            // Old is now a dummy- give it 2 kids , c and newc
            [old setChild: c];
            [c setNext: newc];
        }
        // We've done a merge, so keep the same prev
        c = prev;
    }
    [subjectTable removeAllObjects];                             
}

@end
